import Foundation

extension Array
{
    // 各要素の順列違いを全て生成する。n!個のパターンを作る。nが8を越えると猛烈に遅くなる。
    // 注意：nが大きくなると爆発的に巨大な数になるのでカードゲームのシャッフルなどには使わないこと
    // https://stackoverflow.com/questions/34968470/calculate-all-permutations-of-a-string-in-swift
    public func permutations() -> [[Element]]
    {
        var scratch = Array(self) // This is a scratch space for Heap's algorithm
        var result: [[Element]] = [] // This will accumulate our result
        
        // Heap's algorithm
        func heap(_ n: Int) 
        {
            if n <= 1 
            {
                result.append(scratch)
                
                return
            }
            
            for i in 0..<n-1 
            {
                heap(n-1)
                let j = (n%2 == 1) ? 0 : i
                swap(&scratch[j], &scratch[n-1])
            }
            
            heap(n-1)
        }
        
        // Let's get started
        heap(scratch.count)
        
        // And return the result we built up
        return result
    }
}
