//
//  MazeResolver.swift
//  MazeBoard
//
//  Created by narita on 2017/07/12.
//  Copyright © 2017年 narita. All rights reserved.
//

import Foundation

open class MazeResolver
{
    let mazeBoard : MazeBoard
    
    var route     : Set<MazePos> = []  // 探索している道。マスの集合
    var deadEnd   : Set<MazePos> = []  // 行き止まりの道。マスの集合
    
    public init(_ mazeBoard : MazeBoard)
    {
        self.mazeBoard = mazeBoard
    }
    
    // 返り値は、(true) ゴールだった。(false) 行き止まりだった の2種類。　
    // 引数は、探索する座標を示すMazePosを1個
    public func resolve(_ pos:MazePos, _ closure: (_ m: MazeResolver ) -> Void = { _ in})
    {
        
        func internal_resolve(_ pos:MazePos, _ closure: (_ m: MazeResolver ) -> Void = { _ in})  -> Bool
        {
            closure(self)

            // ゴールならば成功して終了
            if mazeBoard.typeFor( pos: pos ) == .Goal
            {
                return true
            }
            
            for i in findPassages(pos)
            {                 
                self.deadEnd.remove(i)
                self.route.insert(i)

                if internal_resolve(i, closure)
                {    
                    return true
                }
                
                self.route.remove(i)
                self.deadEnd.insert(i)                    
            }
            
            return false
        }
        
        _ = internal_resolve(pos, closure)
    }
    
    
    
    // 与えられた地点から四方を見て、進める方向の配列を返す。
    // 進めるのは、( Digged || Goal ) で、deadEnd(行き止まりへの道)やroute(すでに通過した道)に含まれない物
    public func findPassages(_ pos:MazePos) -> [MazePos]
    {
        var result : Array<MazePos> = []
        
        // 掘削可能な位置を返す
        result = self.roundPositions(pos) // 四方の座標を配列で取得
            .flatMap
            {                    
                // 進入可能な位置だけをフィルタリング
                let type:MazeWallType = mazeBoard.typeFor( pos: $0 )
                
                return ( type == .Digged || type == .Goal ) ? $0 : nil
            } 
            .flatMap
            {                
                ( self.route.contains($0) || self.deadEnd.contains($0) ) ? nil : $0
            }
        
        return result
    }    
    
    // 上下左右の位置の配列を返す。四方
    public func roundPositions(_ pos:MazePos) -> [MazePos]
    {        
        return [MazePos(0, -1), MazePos(0, 1), MazePos(-1, 0), MazePos(1, 0)].map{ pos + $0 }
    }
    
    public func DebugPrint()
    {        
        print(
            (0 ..< self.height) .map
                {
                    i in
                    
                    (0 ..< self.width).map
                        {
                            j in
                            
                            let thePos = MazePos(j, i) 
                            var c = mazeBoard.mazeWallTypeToString( type: mazeBoard.typeFor( pos: thePos ))
                            
                            if self.route.contains(thePos)
                            {
                                c = " ."
                            }
                            else if self.deadEnd.contains(thePos)
                            {
                                c = " x"
                            }
                            else
                            {
                                c = mazeBoard.mazeWallTypeToString( type: mazeBoard.typeFor( pos: thePos ))
                            }
                            
                            return c
                        }.joined() + "\n"
                    
                }.joined()
        )            
    }   
    
    var width : Int
    {
        return self.mazeBoard.width
    }

    var height : Int
    {
        return self.mazeBoard.height
    }    
}

